package CPU.instructions.ROL;

/**
 * The ROL instruction where data to be rotated is in a data reg and the rotate count is immediate.
 * 1110ccc1ss011rrr = 0xe118  
 * 
 * mmm = mode 010
 * rrr = data register to be shifted
 * ccc = the shift count
 * ss =size :   00 = byte
 *              01 = word
 *              10 = long
 * 
 * 
 * @author and copyright for my code (c) Wolfgang Lenerz 2013 - 2020.
 * @version 1.01 speedup
 * 
 * Based on code by Tony Headford, see his licence in accompanying file
 */

import CPU.DisassembledOperand;
import CPU.InstructionSet;
import CPU.DisassembledInstruction;
import CPU.Size;

public class ROLimm implements InstructionSet
{
    private static final int bMask = 0x80;
    private static final int wMask = 0x8000;
    
    @Override
    public final void register(final CPU.MC68000Cpu cpu2)
    {
        int base;
        CPU.Instruction i;
        for(int sz = 0; sz < 3; sz++)                       // byte, word and long only
        {
            switch (sz)
            {
                case 0:                                                         // rotate lsb
                    base = 0xe118;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int reg=opcode&0x7;
                            int nbrRots = ((opcode>>9)&0x7 );                   // how often we rotate
                            int val=cpu.data_regs[reg]&0xff;                    // byte to be rotated
                            if (nbrRots!=0)
                            {
                                val = (val << nbrRots) | (val >> (8 - nbrRots));
                                cpu.data_regs[reg]&=0xffffff00;                 // get rid of ls byte
                                cpu.data_regs[reg]|=val&0xff;                   // and fill it in with new value
                            }
                            cpu.reg_sr&=0xfff0;                                 // 4 status flags set to 0, X unaffected
                            if ((bMask & val)!=0)                               // ms bit set
                                cpu.reg_sr|=8;                                  // so set N 
                            else if (val==0)
                                cpu.reg_sr+=4;                                  // set Z flag
                            if ((val&1)!=0)                                     // if last bit rotated out !=0 : 
                                cpu.reg_sr+=1;                                  // set C flag
     
   /*  old, delete at next review 
                            int nbrShifts = ((opcode>>9)&0x7 );// how often we shift
                            if (nbrShifts==0)
                                nbrShifts=8;
                            int val=cpu.data_regs[opcode&0x7]&0xff;// byte to be shifted
                            int lastOut=0;
                            for (int cnt=0;cnt<nbrShifts;cnt++)
                            {
                                lastOut=val&0x80;                   // highest bit before shift
                                val<<=1;                            // shift one bit up
                                if (lastOut!=0)
                                    val|=1;
                            }
                            val&=0xff;
                            cpu.data_regs[opcode&0x7]&=0xffffff00; //get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;// and fill it in with new value
                            cpu.reg_sr&=0xfff0;                // all status flags set to 0,
                            if ((val&0x80)!=0)                      // ms bit set
                                cpu.reg_sr|=8;                 // so set N flag
                            else if (val==0)
                                cpu.reg_sr+=4;                 // set Z flag
                            if (lastOut!=0)
                                cpu.reg_sr|=1;                // set C flags
*/
                        }
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Byte, cpu);
                        }
                    };  
                    break;
            
                case 1:                                         // word sized rotate
                    base = 0xe158;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int reg=opcode&0x7;
                            int nbrRots = ((opcode>>9)&0x7 );                   // how often we rotate
                            int val=cpu.data_regs[reg]&0xffff;                  // word to be rotated 
                            if (nbrRots==0)
                                nbrRots=8;
                            val = (val << nbrRots) | (val >> (16 - nbrRots));
                            cpu.data_regs[reg]&=0xffff0000;                     // get rid of ls word
                            cpu.data_regs[reg]|=val&0xffff;                     // and fill it in with new value
                            
                            cpu.reg_sr&=0xfff0;                                 // 4 status flags set to 0, X unaffected
                            if ((wMask & val)!=0)                               // ms bit set
                                cpu.reg_sr|=8;                                  // so set N 
                            else if (val==0)
                                cpu.reg_sr+=4;                                  // set Z flag
                            if ((val&1)!=0)                                     // if last bit rotated out !=0 : 
                                cpu.reg_sr+=1;                                  // set C flag
     
       /*      old, delete at next review
                            int nbrShifts = ((opcode>>9)&0x7 );// how often we shift
                            if (nbrShifts==0)
                                nbrShifts=8;
                            int val=cpu.data_regs[opcode&0x7]&0xffff;// word to be shifted
                            int lastOut=0;
                            for (int cnt=0;cnt<nbrShifts;cnt++)
                            {
                                lastOut=val&0x8000;                   // highest bit before shift
                                val<<=1;                            // shift one bit up
                                if (lastOut!=0)
                                    val|=1;
                            }
                            val&=0xffff;
                            cpu.data_regs[opcode&0x7]&=0xffff0000; //get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;// and fill it in with new value
                            cpu.reg_sr&=0xfff0;                // all status flags set to 0,
                            if ((val&0x8000)!=0)                      // ms bit set
                                cpu.reg_sr|=8;                 // so set N flag
                            else if (val==0)
                                cpu.reg_sr+=4;                 // set Z flag
                            if (lastOut!=0)
                                cpu.reg_sr|=1;                // set C flags
*/
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Word, cpu);
                        }
                    };  
                    break;
                    
                default:                                        // long word rotate
                    base = 0xe198;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int reg = opcode&0x7;
                            int nbrRots = ((opcode>>9)&0x7 );   // how often we rotate
                            int val=cpu.data_regs[reg];
                            if (nbrRots==0)
                                nbrRots=8;
                            val=(val << nbrRots) | (val >>> -nbrRots);
                            cpu.data_regs[reg]=val;
                            cpu.reg_sr&=0xfff0;                                 // 4 status flags set to 0, X unaffected
                            if (val<0)                                          // ms bit set
                                cpu.reg_sr|=8;                                  // so set N 
                            else if (val==0)
                                cpu.reg_sr+=4;                                  // set Z flag
                            if ((val&1)!=0)                                     // if last bit rotated out !=0 : 
                                cpu.reg_sr+=1;                                  // set C flag
                            
                            
      /*  old, delete at next review
                            int nbrShifts = ((opcode>>9)&0x7 );     // how often we shift
                            if (nbrShifts==0)
                                nbrShifts=8;
                            int val=cpu.data_regs[opcode&0x7]; // long word to be shifted
                            int lastOut=0;                          // current state of highest bit
                            for (int cnt=0;cnt<nbrShifts;cnt++)
                            {
                                lastOut=val&0x80000000;
                                val<<=1;                            // shift one bit up
                                if (lastOut!=0)
                                    val|=1;
                            }
                            cpu.data_regs[opcode&0x7]=val;// and fill it in with new value
                            cpu.reg_sr&=0xfff0;                // all status flags set to 0,
                            if (val<0)                      // ms bit set
                                cpu.reg_sr|=8;                 // so set N flag
                            else if (val==0)
                                cpu.reg_sr+=4;                 // set Z flag
                            if (lastOut!=0)
                                cpu.reg_sr|=1;                // set C flags
*/
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Long, cpu);
                        }
                    };  break;
            }
            for (int regcnt=0;regcnt<8;regcnt++)
            {
                for (int destreg=0;destreg<8;destreg++)
                {
                    cpu2.addInstruction(base+ (regcnt<<9) + destreg, i);
                }
            }
        }
    }
    protected final DisassembledInstruction disassembleOp(int address, int opcode, Size sz, CPU.MC68000Cpu cpu)
    {
        int count = (opcode >> 9) & 0x07;
        if(count == 0)
                count = 8;
        DisassembledOperand src = new DisassembledOperand("#" + count);
	DisassembledOperand dst = new DisassembledOperand("d" + (opcode & 0x07));
        return new DisassembledInstruction(address, opcode, "rol" + sz.ext(), src, dst);
    }
}
