package CPU.instructions.ROR;

/**
 * The ROR instruction where data to be rotated is in a data reg and the rotate count is in another data reg.
 * 1110ccc0ss111rrr = 0xe038   
 * 
 * mmm = mode 010
 * rrr = data register to berotated
 * ccc = dataregister containing therotate count (modulo 64)
 * ss =size :   00 = byte
 *              01 = word
 *              10 = long
 * 
 * 
 * @author and copyright for my code (c) Wolfgang Lenerz 2014-2020.
 * @version 
 * 1.01 revamped code entirely (correcting bug for sr if rotate count = 0)
 * 
 * Based on code by Tony Headford, see his licence in accompanying file.
 */

import CPU.DisassembledOperand;
import CPU.InstructionSet;
import CPU.DisassembledInstruction;
import CPU.Size;

public class RORreg implements InstructionSet
{
    private static final int bMask = 0x80;
    private static final int wMask = 0x8000;
 //   private static final int lMask = 0x80000000;
        
    @Override
    public final void register(final CPU.MC68000Cpu cpu2)
    {
        int base;
        CPU.Instruction i;
        for (int sz = 0; sz < 3; sz++)                          // byte, word and long
        {
            switch (sz)
            {
                case 0:                                         // rotate lsb
                    base = 0xe038;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int reg = opcode&0x7;
                            int nbrRots = (cpu.data_regs[(opcode>>9)&0x7])&63;  // how often werotate
                            int val=cpu.data_regs[reg]&0xff;                    // byte to be rotated
                            if (nbrRots!=0)
                            {
                                int rots = nbrRots & 7;
                                val =((val>> rots)|(val<<8-rots))&0xff;
                                cpu.data_regs[reg]&=0xffffff00;                 // get rid of ls byte
                                cpu.data_regs[reg]|=val;                        // and fill it in with new value
                            }
                            cpu.reg_sr&=0xfff0;                                 // all status flags set to 0, except for X
                            if ((val &bMask)!=0)                                // ms bit set ?
                            {
                                cpu.reg_sr|=8;                                  // yes, so set N flag
                                if (nbrRots!=0)                                 // C flag is cleared by 0 countrotate
                                    cpu.reg_sr|=1 ;                             // set C flag
                            }
                            else if (val==0)
                                cpu.reg_sr|=4;                 // set Z flag

                            
                            /*      remove at next review
                            
                            int nbrShifts = (cpu.data_regs[(opcode>>9)&0x7])&63;// how often werotate
                            int val=cpu.data_regs[opcode&0x7]&0xff;// byte to berotated
                            int lastOut=0;                              // this is wrong if value<7f,  should be:
                            
                            if (nbrShifts==0)
                            {
                                lastOut = (val & 0x80)!=0?1:0;
                            }
                            else
                                lastOut=0;
                            
                            
                             for (int cnt=0;cnt<nbrShifts;cnt++)
                            {
                                lastOut=val&1;                      // lowest bit beforerotate
                                val>>>=1;                           //rotate one bit down
                                if (lastOut!=0)
                                val|=0x80;                      // rotate bit in
                            }
                            cpu.data_regs[opcode&0x7]&=0xffffff00; //get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;// and fill it in with new value
                            cpu.reg_sr&=0xfff0;                // all status flags set to 0, except for X
                            if (lastOut!=0)                         // ms bit set
                            {
                            cpu.reg_sr|=8;                 // so set N flag if (val==0)
                            if (nbrShifts!=0)                   // C flag is cleared by 0 countrotate
                            cpu.reg_sr|=1 ;            // set C flag
                            }
                            else if (val==0)
                            cpu.reg_sr|=4;                 // set Z flag
                            */
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Byte, cpu);
                        }
                    };  
                    break;
                    
                case 1:                                                 // word sizedrotate
                    base = 0xe078;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int reg = opcode&0x7;
                            int nbrRots = (cpu.data_regs[(opcode>>9)&0x7])&63;  // how often werotate
                            int val=cpu.data_regs[reg]&0xffff;                  // word to berotated
                            if (nbrRots!=0)
                            {
                                int rots = nbrRots & 15;
                                val =((val>> rots)|(val<<16-rots))&0xffff;
                                cpu.data_regs[reg]&=0xffff0000;                 // get rid of ls word
                                cpu.data_regs[reg]|=val;                        // and fill it in with new value
                            }
                            cpu.reg_sr&=0xfff0;                                 // all status flags set to 0, except for X
                            if ((val &wMask)!=0)                                // ms bit set ?
                            {
                                cpu.reg_sr|=8;                                  // yes, so set N flag
                                if (nbrRots!=0)                                 // C flag is cleared by 0 countrotate
                                    cpu.reg_sr|=1 ;                             // set C flag
                            }
                            else if (val==0)
                                cpu.reg_sr|=4;                                  // set Z flag
                            
                          /*  
                            int nbrShifts = (cpu.data_regs[(opcode>>9)&0x7])&63;// how often werotate
                            int val=cpu.data_regs[opcode&0x7]&0xffff;// word to berotated
                            int lastOut=0;                              // !! wrong, see for byte
                            for (int cnt=0;cnt<nbrShifts;cnt++)
                            {
                                lastOut=val&1;                      // lowest bit beforerotate
                                val>>>=1;                           //rotate one bit down
                                if (lastOut!=0)
                                    val|=0x8000;                      // rotate bit in
                            }
                            cpu.data_regs[opcode&0x7]&=0xffff0000; //get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;// and fill it in with new value
                            cpu.reg_sr&=0xfff0;                // all status flags set to 0, except for X
                            if (lastOut!=0)                         // ms bit set
                            {
                                cpu.reg_sr|=8;                 // so set N flag if (val==0)
                                if (nbrShifts!=0)                   // C flag is cleared by 0 countrotate
                                    cpu.reg_sr|=1 ;            // set C flag
                            }
                            else if (val==0)
                                cpu.reg_sr|=4;                 // set Z flag
*/
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Word, cpu);
                        }
                        
                    };  
                    break;
                    
                default:
                    base = 0xe0b8;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int reg = opcode&0x7;
                            int nbrRots = (cpu.data_regs[(opcode>>9)&0x7])&63;  // how often werotate
                            int val=cpu.data_regs[reg];                         // lword to berotated
                            if (nbrRots!=0)
                            {
                                int rots = nbrRots & 31;
                                val =(val>>> rots)|(val<<32-rots);
                                cpu.data_regs[reg] =val;                        // and fill it in with new value
                            }
                            cpu.reg_sr&=0xfff0;                                 // all status flags set to 0, except for X
                            if (val<0)                                          // ms bit set ?
                            {
                                cpu.reg_sr|=8;                                  // yes, so set N flag
                                if (nbrRots!=0)                                 // C flag is cleared by 0 countrotate
                                    cpu.reg_sr|=1 ;                             // set C flag
                            }
                            else if (val==0)
                                cpu.reg_sr|=4;                                  // set Z flag
           /*                 
                            int nbrShifts = (cpu.data_regs[(opcode>>9)&0x7])&63;// how often werotate
                            int val=cpu.data_regs[opcode&0x7]; // long word to berotated
                            int lastOut=0;                          // current state of lowest bit
                            for (int cnt=0;cnt<nbrShifts;cnt++)
                            {
                                lastOut=val&1;
                                val>>>=1;                            //rotate one bit down with sign extension
                                if (lastOut!=0)
                                    val|=0x80000000;                      // rotate bit in
                            }
                            cpu.data_regs[opcode&0x7]=val;     // and fill it in with new value
                            cpu.reg_sr&=0xfff0;                // all status flags set to 0, except for X
                            if (lastOut!=0)                         // ms bit set
                            {
                                cpu.reg_sr|=8;                 // so set N flag if (val==0)
                                if (nbrShifts!=0)                   // C flag is cleared by 0 countrotate
                                    cpu.reg_sr|=1 ;            // set C flag
                            }
                            else if (val==0)
                                cpu.reg_sr|=4;                 // set Z flag
*/
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Long, cpu);
                        }
                    };  
                    break;
            }
            for (int regcnt=0;regcnt<8;regcnt++)
            {
                for (int destreg=0;destreg<8;destreg++)
                {
                    cpu2.addInstruction(base+ (regcnt<<9) + destreg, i);
                }
            }
        }
    }
    protected final DisassembledInstruction disassembleOp(int address, int opcode, Size sz, CPU.MC68000Cpu cpu)
    {
           //reg mode
            DisassembledOperand src = new DisassembledOperand("d" + ((opcode >> 9) & 0x07));
            DisassembledOperand dst = new DisassembledOperand("d" + (opcode & 0x07));
            return new DisassembledInstruction(address, opcode, "ror" + sz.ext(), src, dst);
    }
}   
    
