package CPU.instructions.ROX;

/**
 * The ROXL instruction where data to be rotated is in a data reg and the rotate count is immediate.
 * 1110ccc1ss010rrr = 0xe110 
 * 
 * rrr = data register to be shifted
 * ccc = the shift count
 * ss =size :   00 = byte
 *              01 = word
 *              10 = long
 * 
 * 
 * 
 * 
 * @author and copyright for my code (c) Wolfgang Lenerz 2014 -2020.
 * @version 1.01 correct handling of X flag, remove barrel shift.
 * 
 * Based on code by Tony Headford, see his licence in accompanying file
 */

import CPU.DisassembledOperand;
import CPU.InstructionSet;
import CPU.DisassembledInstruction;
import CPU.Size;

public class ROXLimm implements InstructionSet
{
    @Override
    public final void register(final CPU.MC68000Cpu cpu2)
    {
        int base;
        CPU.Instruction i;
        for(int sz = 0; sz < 3; sz++)                           // byte, word and long
        {
            switch (sz)
            {
                case 0:                                         // rotate lsb
                    base = 0xe110;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)     
                        {
                            int nbrShifts = ((opcode>>9)&0x7 );     // how often we rotate
                            int val=cpu.data_regs[opcode&0x7]&0xff; // byte to be rotated
                            int xBit=(cpu.reg_sr &16)==0?0:0x80;
                            int cxflag;
                            
                            if (nbrShifts ==0)                      // 0 = rotate by 8 -> full rotate, easy
                            {
                                cxflag = val & 1;
                                val>>>=1;
                                val+=xBit;
                            }
                            else
                            {
                                int temp = 0x80 >>>(nbrShifts -1);
                                cxflag = val & temp;                // identify bit setting c + x flag         
                                temp = val>>>1;                     
                                temp+=xBit;
                                val =(val<<nbrShifts) | (temp>>>(8-nbrShifts));
                            }
                            val&=0xff;
                            cpu.data_regs[opcode&0x7]&=0xffffff00;  //get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;         // and fill it in with new value
                            cpu.reg_sr&=0xffe0;                     // all status flags set to 0 (there is no 0 rotate count here)
                            if (val==0)
                                cpu.reg_sr|=4;                      // set Z flag
                            else if ((val&0x80)!=0)                 // ms bit set
                                cpu.reg_sr|=8;                      // so set N flag
                            if (cxflag!=0)
                                cpu.reg_sr|=17;                     // set C & X flags
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)     
                        {
                            return disassembleOp(address, opcode, Size.Byte, cpu);
                        };
                    }; 
                    break;
            
                case 1:                                             // word sized shift
                    base = 0xe150;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)     
                        {
                            int nbrShifts = ((opcode>>9)&0x7 ); // how often we rotate
                            int val=cpu.data_regs[opcode&0x7]&0xffff; // word to be rotated 
                            if (nbrShifts==0)
                                nbrShifts=8;                    // min rotate count is 1
                            int xBit=((cpu.reg_sr &16)==0)?0:0x8000;
                            int temp = 0x8000 >>>(nbrShifts -1);
                            boolean cxflag = (val & temp)!=0;                 
                            temp = val>>>1;                     
                            temp+=xBit;
                            val =(val<<nbrShifts) | (temp>>>(16-nbrShifts)); //(*d)=((*d)<<c)|(temp>>(8-c));
                            val &=0xffff;
                            cpu.data_regs[opcode&0x7]&=0xffff0000;  //get rid of ls word
                            cpu.data_regs[opcode&0x7]|=val;         // and fill it in with new value
                            cpu.reg_sr&=0xffe0;                     // all status flags set to 0 (there is no 0 rotate count here)
                            if (val==0)
                                cpu.reg_sr|=4;                      // set Z flag
                            else if ((val&0x8000)!=0)                 // ms bit set
                                cpu.reg_sr|=8;                      // so set N flag
                            if (cxflag)
                                cpu.reg_sr|=17;
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Word, cpu);
                        }
                    };  
                    break;
                    
                default:                                            // rotate long word
                    base = 0xe190;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int nbrShifts = ((opcode>>9)&0x7 );     // how often we rotate
                            if (nbrShifts==0)
                                nbrShifts=8;                        // min rotate count is 1
                            int val=cpu.data_regs[opcode&0x7];      // lw to be rotated
                            int xBit=((cpu.reg_sr &16)==0)?0:0x80000000;
                            int temp = 0x80000000 >>>(nbrShifts -1);
                            boolean cxflag = (val & temp)!=0;                 
                            temp = val>>>1;                     
                            temp+=xBit;
                            val =(val<<nbrShifts) | (temp>>>(32-nbrShifts)); 
                            cpu.data_regs[opcode&0x7]=val;          // and fill it in with new value
                            cpu.reg_sr&=0xffe0;                     // all status flags set to 0 (there is no 0 rotate count here)
                            if (val==0)
                                cpu.reg_sr|=4;                      // set Z flag
                            else if (val<0)               // ms bit set
                                cpu.reg_sr|=8;                      // so set N flag
                            if (cxflag)
                                cpu.reg_sr|=17;
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Long, cpu);
                        }
                    };  
                    break;
            }
            for (int regcnt=0;regcnt<8;regcnt++)
            {
                for (int destreg=0;destreg<8;destreg++)
                {
                    cpu2.addInstruction(base+ (regcnt<<9) + destreg, i);
                }
            }
        }
    }
    protected final DisassembledInstruction disassembleOp(int address, int opcode, Size sz, CPU.MC68000Cpu cpu)
    {
        int count = (opcode >> 9) & 0x07;
        if(count == 0)
                count = 8;
        DisassembledOperand src = new DisassembledOperand("#" + count);
	DisassembledOperand dst = new DisassembledOperand("d" + (opcode & 0x07));
        return new DisassembledInstruction(address, opcode, "roxl" + sz.ext(), src, dst);
    }
}
