package CPU.instructions.ROX;

/**
 * The ROXR instruction where data to be rotated is in a data reg and the rotate count is immediate.
 * 1110ccc0ss010rrr = 0xe010   
 * 
 * mmm = mode 010
 * rrr = data register to be shifted
 * ccc = the shift count
 * ss =size :   00 = byte
 *              01 = word
 *              10 = long
 * 
 * @author and copyright for my code (c) Wolfgang Lenerz 2014.
 * @version 1.01 correct handling of X flag, remove barrel shift.
 * 
 * Based on code by Tony Headford, see his licence in accompanying file.
 */

import CPU.DisassembledOperand;
import CPU.InstructionSet;
import CPU.DisassembledInstruction;
import CPU.Size;

public class ROXRimm implements InstructionSet
{
    @Override
    public final void register(final CPU.MC68000Cpu cpu2)
    {
        int base;
        CPU.Instruction i;
        for(int sz = 0; sz < 3; sz++)                       // byte, word and long only
        {
            switch (sz)                                         // rotate byte
            {
                case 0:
                    base = 0xe010;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int nbrRotats = ((opcode>>9)&0x7 ); // how often we rotate
                            if (nbrRotats==0)
                                nbrRotats=8;
                            int val=cpu.data_regs[opcode&0x7]&0xff;// byte to be rotated
                            int xBit=((cpu.reg_sr &16)==0)?0:1;
                            int temp = 1<<(nbrRotats -1);
                            boolean cxflag = (val & temp)!=0;   // identify bit setting c + x flag         
                            temp = val<<1;                     
                            temp|=xBit;
                            val =(val>>>nbrRotats) | (temp<<(8-nbrRotats));
                            val&=0xff;
                            cpu.data_regs[opcode&0x7]&=0xffffff00; //get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;     // and fill it in with new value
                            cpu.reg_sr&=0xffe0;                 // status flags set to 0, X can't be unaffected : no 0 shift count here
                            if (val==0)
                                cpu.reg_sr|=4;                  // set Z flag
                            else if ((val&0x80)!=0)             // ms bit set
                                cpu.reg_sr|=8;                  // so set N flag 
                            if (cxflag)
                                cpu.reg_sr|=17;                 // set C & X flags
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Byte, cpu);
                        }
                    };  
                    break;
           
                case 1:                                         // rotate word
                    base = 0xe050;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int nbrRotats = ((opcode>>9)&0x7 ); // how often we rotate
                            if (nbrRotats==0)
                                nbrRotats=8;
                            int val=cpu.data_regs[opcode&0x7]&0xffff;// byte to be rotated
                            int xBit=((cpu.reg_sr &16)==0)?0:1;
                            int temp = 1<<(nbrRotats -1);
                            boolean cxflag = (val & temp)!=0;   // identify bit setting c + x flag         
                            temp = val<<1;                     
                            temp|=xBit;
                            val =(val>>>nbrRotats) | (temp<<(16-nbrRotats));
                            val&=0xffff;
                            cpu.data_regs[opcode&0x7]&=0xffff0000; //get rid of ls byte
                            cpu.data_regs[opcode&0x7]|=val;     // and fill it in with new value
                            cpu.reg_sr&=0xffe0;                 // status flags set to 0, X can't be unaffected : no 0 shift count here
                            if (val==0)
                                cpu.reg_sr|=4;                  // set Z flag
                            else if ((val&0x8000)!=0)             // ms bit set
                                cpu.reg_sr|=8;                  // so set N flag 
                            if (cxflag)
                                cpu.reg_sr|=17;                 // set C & X flags
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Word, cpu);
                        }
                    };  
                    break;
                default:                                        // rotate LW
                    base = 0xe090;
                    i = new CPU.Instruction()
                    {
                        @Override
                        public final void execute(int opcode,CPU.MC68000Cpu cpu)
                        {
                            int nbrRotats = ((opcode>>9)&0x7 );  // how often we rotate
                            if (nbrRotats==0)
                                nbrRotats=8;
                            int val=cpu.data_regs[opcode&0x7];  // byte to be rotated
                            int xBit=((cpu.reg_sr &16)==0)?0:1;
                            int temp = 1<<(nbrRotats -1);
                            boolean cxflag = (val & temp)!=0;   // identify bit setting c + x flag         
                            temp = val<<1;                     
                            temp|=xBit;
                            val =(val>>>nbrRotats) | (temp<<(32-nbrRotats));
                            cpu.data_regs[opcode&0x7]=val;      //new value
                            cpu.reg_sr&=0xffe0;                 // status flags set to 0, X can't be unaffected : no 0 shift count here
                            if (val==0)
                                cpu.reg_sr|=4;                  // set Z flag
                            else if (val<0)                     // ms bit set
                                cpu.reg_sr|=8;                  // so set N flag 
                            if (cxflag)
                                cpu.reg_sr|=17;                 // set C & X flags
/*
                             int nbrRotats = ((opcode>>9)&0x7 );     // how often we shift
                            if (nbrRotats==0)
                                nbrRotats=8;
                            int val=cpu.data_regs[opcode&0x7]; // long word to be shifted
                            int lastOut=val&0x1;
                            int xBit=((cpu.reg_sr &16)==0)?0:1;
                            val>>>=1;
                            if (xBit!=0)
                                val|=0x80000000;                    // set x bit
                            for (int cnt=1;cnt<nbrRotats;cnt++)     // WATCH the start count : 1, not 0
                            {
                                lastOut=val&0x1;                    // lowest bit before shift
                                val>>>=1;                           // shift one bit up
                                if (lastOut!=0)
                                    val|=0x80000000;                // set
                            }
                            cpu.data_regs[opcode&0x7]=val;     // and fill it in with new value
                            cpu.reg_sr&=0xffe0;                // status flags set to 0, X can't be unaffected : no 0 shift count here
                            if (val==0)
                                cpu.reg_sr|=4;                 // set Z flag
                            else if (val<0)                         // ms bit set
                                cpu.reg_sr|=8;                 // so set N flag (this may be set if the X dlag was st, even if last bit rotatd out wand't)
                            if (lastOut!=0)
                                cpu.reg_sr|=17;                // so set C + X flags
*/
                        }
                        
                        @Override
                        public DisassembledInstruction disassemble(int address, int opcode,CPU.MC68000Cpu cpu)
                        {
                            return disassembleOp(address, opcode, Size.Long, cpu);
                        }
                    };  break;
            }
            for (int regcnt=0;regcnt<8;regcnt++)
            {
                for (int destreg=0;destreg<8;destreg++)
                {
                    cpu2.addInstruction(base+ (regcnt<<9) + destreg, i);
                }
            }
        }
    }
    
    protected final DisassembledInstruction disassembleOp(int address, int opcode, Size sz, CPU.MC68000Cpu cpu)
    {
        int count = (opcode >> 9) & 0x07;
        if(count == 0)
                count = 8;
        DisassembledOperand src = new DisassembledOperand("#" + count);
	DisassembledOperand dst = new DisassembledOperand("d" + (opcode & 0x07));
        return new DisassembledInstruction(address, opcode, "roxr" + sz.ext(), src, dst);
    }
}
